/*
  Microwriter training tool demo for pro micro.
  Microwriter has 6 buttons.
  Uses 4*20 LCD display via I2C.
  Setup is for righ handed person only.
  October 2021.
*/

//#define my_test //In test mode 

#include <Wire.h> 
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x3F,20,4);// Set the LCD I2C address here 0x3F in my case but usually 0x27

// make some custom characters:
byte char_box[8] = {
  0b11111,
  0b10001,
  0b10001,
  0b10001,
  0b10001,
  0b10001,
  0b10001,
  0b11111
};
byte char_full_box[8] = {
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111,
  0b11111
};


const int box=0;
const int full_box=1;

#define button_1 10 //thumb control
#define button_2 16 //thumb
#define button_3 6  //index
#define button_4 7  //middle
#define button_5 8  //ring
#define button_6 9  //pinkie

//required to debounce switch
#define MAX_CHECKS 10
volatile uint8_t Debounced_State=0;//accessed by isr and main loop code
uint8_t State[MAX_CHECKS]={0};
uint8_t Index=0;

// the setup function runs once when you press reset or power the board
void setup() {

  pinMode(button_1,INPUT_PULLUP);
  pinMode(button_2,INPUT_PULLUP);
  pinMode(button_3,INPUT_PULLUP);
  pinMode(button_4,INPUT_PULLUP);
  pinMode(button_5,INPUT_PULLUP);
  pinMode(button_6,INPUT_PULLUP);

    // initialize timer1 
  noInterrupts();           // disable all interrupts
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1  = 0;

  OCR1A = 625;              // compare match register  16MHX/256/100HZ 
  TCCR1B |= (1 << WGM12);   // CTC mode
  TCCR1B |= (1 << CS12);    // 256 prescaler 
  TIMSK1 |= (1 << OCIE1A);  // enable timer compare interrupt
  interrupts();             // enable all interrupts

//-------- Turn on display and write characters ------------------  
  
  lcd.init();// initialize the lcd 
  lcd.init();
  lcd.backlight();
  
  // create a new characters
  lcd.createChar(box, char_box);
  lcd.createChar(full_box, char_full_box);

  reset_fingers();
  
}


#ifdef my_test
//for test purposes
char characters_array[]=",-. @&%;:/!?=X+*$()";// minus \ " £
const int characters_array_elements= 19;
const int characters_array_max_index= characters_array_elements-1;
#else
//full character set
char characters_array[]="abcdefghijklmnopqrstuvwxyz0123456789,-. @&%;:/!?=X+*$()ABCDEFGHIJKLMNOPQRSTUVWXYZ";// minus \ " £
const int characters_array_elements= 81;
const int characters_array_max_index= characters_array_elements-1;
#endif


// the loop function runs over and over again forever
void loop() {

static bool again=true;
static bool do_reset=false;

static int index=characters_array_max_index;
static int score=0;

static bool do_character=false;
static bool button_up=false;
static bool chord_ready=false;
static uint8_t stored_state=0;
static uint8_t decode_me=0;
static char print_me;

static bool numeric_shift=false;
static int numeric_shift_index=0;

static bool upper_case=false;
static int upper_case_index=0;

if(again){fisher_yates_shuffle(characters_array,characters_array_max_index);again=false;lcd.setCursor(13,0); lcd.write(characters_array[index]);}


    noInterrupts();           // disable all interrupts    
if (Debounced_State>stored_state){stored_state=Debounced_State;chord_ready=true;}//store chord and flag as ready
if (Debounced_State<stored_state){if(chord_ready){button_up=true;decode_me=stored_state;}stored_state=Debounced_State;}    
    interrupts();             // enable all interrupts

if(button_up && chord_ready)
{
chord_ready=false;
button_up=false;

/*
 * Microwriter commands
 * Command key once = single uppercase
 * Command key twice = continuous uppercase, 110000 to go back to lowercase (the two thumb buttons)
 * Backspace 100010
 * Numerics command once for a single number twice to lock into numbers mode, 110000 to go back to lowercase (the two thumb buttons)
*/


if (decode_me==B00100000){upper_case=true;upper_case_index++;}//Uppercase
if (upper_case_index>2){upper_case_index=2;}
if (decode_me==B00110000 && upper_case){upper_case_index=0;upper_case=false;}//two thumb buttons - cancells uppercase lock

if (decode_me==B00100110){numeric_shift=true;numeric_shift_index++;}//Numeric shift
if (numeric_shift_index>2){numeric_shift_index=2;}
if (decode_me==B00110000 && numeric_shift){numeric_shift_index=0;numeric_shift=false;}//two thumb buttons - cancels numbers lock

if(numeric_shift){
   //Numbers
   if (decode_me==B00001000){print_me='1';do_character=true;}
   if (decode_me==B00011000){print_me='2';do_character=true;}                 
   if (decode_me==B00011100){print_me='3';do_character=true;}                  
   if (decode_me==B00011110){print_me='4';do_character=true;}
   if (decode_me==B00011111){print_me='5';do_character=true;}
   if (decode_me==B00000001){print_me='6';do_character=true;}
   if (decode_me==B00000011){print_me='7';do_character=true;}
   if (decode_me==B00000111){print_me='8';do_character=true;}
   if (decode_me==B00001111){print_me='9';do_character=true;}
   if (decode_me==B00000100){print_me='0';do_character=true;}

   //punctuation   
   if (decode_me==B00010100){print_me='(';do_character=true;}
   if (decode_me==B00000101){print_me=')';do_character=true;}
   if (decode_me==B00011001){print_me='£';do_character=true;}//prints other character on acutal machine # if i recall correctly
   if (decode_me==B00000010){print_me='$';do_character=true;}
   if (decode_me==B00001100){print_me='*';do_character=true;}//Asterisk
   if (decode_me==B00001010){print_me='+';do_character=true;}                   
   if (decode_me==B00010001){print_me='-';do_character=true;}                  
   if (decode_me==B00010111){print_me='X';do_character=true;}//"Times" for multiplication - may need to alter to another asterisk                 
   if (decode_me==B00010011){print_me='=';do_character=true;}
   if (decode_me==B00010110){print_me='?';do_character=true;}
   if (decode_me==B00001101){print_me='!';do_character=true;}
   if (decode_me==B00010010){print_me='/';do_character=true;}
   if (decode_me==B00011011){print_me=':';do_character=true;}
   if (decode_me==B00011010){print_me=';';do_character=true;}
   if (decode_me==B00010101){print_me='%';do_character=true;}
   if (decode_me==B00000110){print_me='"';do_character=true;}
   if (decode_me==B00001001){print_me='&';do_character=true;}                  
   if (decode_me==B00011101){print_me='@';do_character=true;}                  
   if (decode_me==B00010000){print_me=' ';do_character=true;}
   if (decode_me==B00001110){print_me='.';do_character=true;}
   if (decode_me==B00001011){print_me=',';do_character=true;}   
}
else
{   
   //Alphabet
   if (decode_me==B00001100){print_me='a';do_character=true;}
   if (decode_me==B00000111){print_me='b';do_character=true;}                 
   if (decode_me==B00010100){print_me='c';do_character=true;}                  
   if (decode_me==B00011100){print_me='d';do_character=true;}
   if (decode_me==B00001000){print_me='e';do_character=true;}
   if (decode_me==B00011110){print_me='f';do_character=true;}
   if (decode_me==B00000011){print_me='g';do_character=true;}
   if (decode_me==B00010001){print_me='h';do_character=true;}
   if (decode_me==B00011000){print_me='i';do_character=true;}
   if (decode_me==B00010011){print_me='j';do_character=true;}
   if (decode_me==B00010010){print_me='k';do_character=true;}
   if (decode_me==B00011001){print_me='l';do_character=true;}        
   if (decode_me==B00001111){print_me='m';do_character=true;}                
   if (decode_me==B00000110){print_me='n';do_character=true;}
   if (decode_me==B00000100){print_me='o';do_character=true;}
   if (decode_me==B00011111){print_me='p';do_character=true;}
   if (decode_me==B00000101){print_me='q';do_character=true;}
   if (decode_me==B00011010){print_me='r';do_character=true;}
   if (decode_me==B00000010){print_me='s';do_character=true;}
   if (decode_me==B00001010){print_me='t';do_character=true;} 
   if (decode_me==B00000001){print_me='u';do_character=true;}
   if (decode_me==B00001001){print_me='v';do_character=true;}       
   if (decode_me==B00011011){print_me='w';do_character=true;}                 
   if (decode_me==B00010111){print_me='x';do_character=true;}
   if (decode_me==B00010110){print_me='y';do_character=true;}
   if (decode_me==B00010101){print_me='z';do_character=true;}
     
   if (decode_me==B00010000){print_me=' ';do_character=true;}//Space                  
   if (decode_me==B00001110){print_me='.';do_character=true;}//Full stop                  
   if (decode_me==B00001101){print_me='-';do_character=true;}//Hyphen
   if (decode_me==B00001011){print_me=',';do_character=true;}//Comma 
   if (decode_me==B00011101){print_me='\'';do_character=true;}//Apostrophe - \ needed as escape character to ensure ' prints correctly
   
}
      
  if (decode_me==B01000100){print_me='\b';do_character=true;}//Backspace character
           
}

if (do_character) 
{
  do_character=false;
  
  if(numeric_shift_index==1){numeric_shift_index=0;numeric_shift=false;}
  if (upper_case && (print_me >= 97 && print_me <=122) ){print_me-=32;}
  if(upper_case_index==1){upper_case_index=0;upper_case=false;}

  //test code - seems to work ok
  if(decode_me==B00010000)  //space symbol
  {set_fingers(decode_me);lcd.setCursor(12,1);lcd.write('\"');lcd.write(' ');lcd.write('\"');delay(1000);lcd.setCursor(12,1);lcd.write(' ');lcd.write(' ');lcd.write(' ');reset_fingers();} 
  else  
  {set_fingers(decode_me);lcd.setCursor(13,1);lcd.write(print_me);delay(1000);lcd.setCursor(13,1);lcd.write(' ');reset_fingers();} 

  if(print_me==characters_array[index])
    {index--; score++;lcd.setCursor(13,3);lcd.print(score);
      if(index<0){lcd.setCursor(11,0);lcd.print("Again?");do_reset=true;}
      else{lcd.setCursor(13,0); lcd.write(characters_array[index]);}
      }
      
   if(print_me=='y'&& do_reset==true)
   {score=0;
    index=characters_array_max_index;
    lcd.setCursor(11,0);
    lcd.print("      ");
    lcd.setCursor(11,1);
    lcd.print("     ");
    lcd.setCursor(11,3);
    lcd.print("     ");
    do_reset=false;
    again=true;
   }
   
}


}

//my_functions

/*
 * Fisher-Yates shuffle.
 * Durstenfeld version - shuffle array items in place.
 * Source - https://en.wikipedia.org/wiki/Fisher%E2%80%93Yates_shuffle - retrieved 22 Febuary 2019
 * 
 * -- To shuffle an array a of n elements (indices 0..n-1):
     for i from n−1 downto 1 do
     j ← random integer such that 0 ≤ j ≤ i
     exchange a[j] and a[i]
*/
void fisher_yates_shuffle(char characters_array[], int characters_array_max_index)
{
  int i=0;
  int j=0;
  int temp=0;
for (i=characters_array_max_index;i>1;i--){
     //j ← random integer such that 0 ≤ j ≤ i
     j=random(0,i+1);
     //swap j->i and i->j
     temp=characters_array[i];
     characters_array[i]=characters_array[j];
     characters_array[j]=temp;
     } 
  }

void reset_fingers()
{
// NOTE: Cursor Position: (CHAR, LINE) start at 0  
  lcd.setCursor(2,0); 
  lcd.write(byte(box));// when calling lcd.write() '0' must be cast as a byte
  
  lcd.setCursor(1,1);
  lcd.write(byte(box));
  lcd.setCursor(3,1);
  lcd.write(byte(box));

  lcd.setCursor(0,2);
  lcd.write(byte(box)); 
  lcd.setCursor(4,2);
  lcd.write(byte(box));

  lcd.setCursor(0,3);
  lcd.write(byte(box));
   
  }
  
void set_fingers(byte fingers)
{ 
  
  
  if(fingers & B00010000)//third row left
  {lcd.setCursor(0,2);
  lcd.write(byte(full_box));}// when calling lcd.write() '0' must be cast as a byte
  
  if(fingers & B00001000)//second row left
  {lcd.setCursor(1,1);
  lcd.write(byte(full_box));}
  
  if(fingers & B0000100)//top
  {lcd.setCursor(2,0); 
  lcd.write(byte(full_box));}
  
  if(fingers & B00000010)//second row right
  {lcd.setCursor(3,1);
  lcd.write(byte(full_box));}
  
  if(fingers & B00000001)//third row right
  {lcd.setCursor(4,2);
  lcd.write(byte(full_box));} 
  
  if(fingers & B00100000){lcd.setCursor(0,3);//fourth row left - command key
  lcd.write(byte(full_box));}
  
  
  }

 
ISR(TIMER1_COMPA_vect)          // timer compare interrupt service routine
{

//read buttons
  uint8_t temp=0x00;
  temp|=!digitalRead(button_1)<<5;
  temp|=!digitalRead(button_2)<<4;
  temp|=!digitalRead(button_3)<<3;
  temp|=!digitalRead(button_4)<<2;
  temp|=!digitalRead(button_5)<<1;
  temp|=!digitalRead(button_6)<<0;
          
//debounce
uint8_t Previous_Debounced_State=Debounced_State;
   
  uint8_t i,j;
  State[Index]= temp;
  ++Index;
  j=0xFF;
  for (i=0;i<MAX_CHECKS;i++){j=j&State[i];}
  Debounced_State=j;
  if(Index>=MAX_CHECKS){Index=0;} 
    
}
//-----------end------------  
